#include "msr.h"
#include "../debug/debug.h"
#include <linux/printk.h>
#include <linux/types.h>

bool ia32_feature_control_flags(void) {
    int msr_value = __rdmsr(IA32_FEATURE_CONTROL_MSR);
    DEBUG_FMT("IA32_FEATURE_CONTROL value = %llu\n",
              __rdmsr(IA32_FEATURE_CONTROL_MSR));

    if (!(msr_value & IA32_FEATURE_CONTROL_LOCK_BIT)) {
        DEBUG_FMT("Lock bit is not set in IA32_FEATURE_CONTROL_MSR\n");

        __wrmsr(IA32_FEATURE_CONTROL_MSR,
                IA32_FEATURE_CONTROL_LOCK_BIT |
                    IA32_FEATURE_CONTROL_VMXON_OUTSIDE_SMX,
                0);
    }

    if (!(msr_value & IA32_FEATURE_CONTROL_VMXON_OUTSIDE_SMX)) {
        return false;
    }

    return true;
}

void patch_control_registers(void) {
    unsigned long long cr0, cr4;

    __asm__ volatile("mov %%cr0, %0" : "=r"(cr0));
    __asm__ volatile("mov %%cr4, %0" : "=r"(cr4));

    unsigned long long cr0_fixed0 = __rdmsr(IA32_VMX_CR0_FIXED0);
    unsigned long long cr0_fixed1 = __rdmsr(IA32_VMX_CR0_FIXED1);

    unsigned long long cr4_fixed0 = __rdmsr(IA32_VMX_CR4_FIXED0);
    unsigned long long cr4_fixed1 = __rdmsr(IA32_VMX_CR4_FIXED1);

    DEBUG_FMT("Pre-patch: CR0 = 0x%llx\n", cr0);
    DEBUG_FMT("Pre-patch: CR4 = 0x%llx\n", cr4);

    // If bit X is 1 in IA32_VMX_CR0_FIXED0, then that bit of CR0 is fixed to 1
    // in VMX operation. Similarly, if bit X is 0 in IA32_VMX_CR0_FIXED1, then
    // that bit of CR0 is fixed to 0 in VMX operation. It is always the case

    // Setting mandatory bits of FIXED0 in CR0
    cr0 |= cr0_fixed0;
    // Clearing mandatory bits of FIXED1 in CR0;
    cr0 &= cr0_fixed1;

    cr4 |= cr4_fixed0;
    cr4 &= cr4_fixed1;

    DEBUG_FMT("Post-patch: CR0 = 0x%llx\n", cr0);
    DEBUG_FMT("Post-patch: CR4 = 0x%llx\n", cr4);

    __asm__ volatile("mov %0, %%cr0" ::"r"(cr0));
    __asm__ volatile("mov %0, %%cr4" ::"r"(cr4));
}
